searchData={"items":[{"type":"module","title":"EXLA","doc":"[Google's XLA](https://www.tensorflow.org/xla/) (Accelerated Linear Algebra) compiler/backend for Nx.\n\nIt supports just-in-time (JIT) compilation to GPU (both CUDA and ROCm) and TPUs.","ref":"EXLA.html"},{"type":"module","title":"XLA binaries - EXLA","doc":"EXLA relies on the [XLA](https://github.com/elixir-nx/xla) package to\nprovide the necessary XLA binaries. Whenever possible it tries to download\nprecompiled builds, but you may need to build from source if there is no\nversion matching your target environment. For more details, including\nGPU/TPU support see [the usage section](https://github.com/elixir-nx/xla#usage).","ref":"EXLA.html#module-xla-binaries"},{"type":"module","title":"Configuration - EXLA","doc":"EXLA ships with a backend to store tensors and run computations on.\nGenerally speaking, the backend is enabled globally in your `config/config.exs`\n(or `config/ENV.exs`) with the following:\n\n    import Config\n    config :nx, :default_backend, EXLA.Backend\n\nIn a script/notebook, you would do:\n\n    Mix.install([\n      {:exla, \"~> 0.2\"}\n    ])\n\n    Nx.global_default_backend(EXLA.Backend)\n\nFrom now on, all created tensors will be allocated directly on the given\n`EXLA.Backend`. You can use functions such as `Nx.backend_transfer/2` to\nexplicitly transfer tensors.\n\nEXLA will pick an available client to allocate and compute tensors, in this\norder: `:cuda`, `:rocm`, `:tpu`, and `:host` (CPU). See the \"Clients\" section\nbelow for more information.\n\nTo use GPUs/TPUs, you must also set the appropriate value for the\n[`XLA_TARGET`](https://github.com/elixir-nx/xla#xla_target) environment\nvariable. If you have GPU/TPU enabled, we recommend setting the environment\nvariable for your machine altogether. For CUDA, setting\n`ELIXIR_ERL_OPTIONS=\"+sssdio 128\"` is also required on more complex operations\nto increase CUDA's compiler stack size.\n\nNote that setting the `EXLA.Backend` does not enable the EXLA compiler.\nYou must still pass the `compiler: EXLA` option to `Nx.Defn` functions\nor call the functions in this module.","ref":"EXLA.html#module-configuration"},{"type":"module","title":"Options - EXLA","doc":"The options accepted by EXLA backend/compiler are:\n\n  * `:client` - an atom representing the client to use. The default\n    client is chosen on this order: `:cuda`, `:rocm`, `:tpu`, and `:host`.\n\n  * `:device_id` - the default device id to run the computation\n      on. Defaults to the `:default_device_id` on the client\n\n  * `:precision` - control the tradeoff between speed and accuracy for\n    array computations on accelerator backends (i.e. TPU and GPU).\n    It must be one of:\n\n    * `:default` - Fastest mode, but least accurate. Performs computations\n      in bfloat16\n\n    * `:high` - Slower but more accurate. Performs float32 computations in\n      3 bfloat16 passes, or using tensorfloat32 where available\n\n    * `:highest` - Slowest but most accurate. Performs computations in float32\n      or float64 as applicable","ref":"EXLA.html#module-options"},{"type":"module","title":"Clients - EXLA","doc":"The `EXLA` library uses a client for compiling and executing code.\nThose clients are typically bound to a platform, such as CPU or\nGPU.\n\nThose clients are singleton resources on Google's XLA library,\ntherefore they are treated as a singleton resource on this library\ntoo. EXLA ships with runtime client configuration for each supported\nplatform:\n\n    config :exla, :clients,\n      cuda: [platform: :cuda],\n      rocm: [platform: :rocm],\n      tpu: [platform: :tpu],\n      host: [platform: :host]\n\nIn a script/notebook, you can set those after `Mix.install/2`,\nbut before any tensor operation is performed:\n\n    Application.put_env(:exla, :clients,\n      cuda: [platform: :cuda],\n      rocm: [platform: :rocm],\n      tpu: [platform: :tpu],\n      host: [platform: :host]\n    )\n\nYou can provide your own list of clients, replacing the list above\nor configuring each client as listed below. You can also specify\n`:default_client` to set a particular client by default or\n`:preferred_clients` to change the order of clients preference,\nbut those configurations are rarely set in practice.\n\n> **Important!** you should avoid using multiple clients for the\n> same platform. If you have multiple clients per platform, they\n> can race each other and fight for resources, such as memory.\n> Therefore, we recommend developers to stick with the default\n> clients above.","ref":"EXLA.html#module-clients"},{"type":"module","title":"Client options - EXLA","doc":"Each client configuration accepts the following options:\n\n  * `:platform` - the platform the client runs on. It can be\n    `:host` (CPU), `:cuda`, `:rocm`, or `:tpu`. Defaults to `:host`.\n\n  * `:default_device_id` - the default device ID to run on.\n    For example, if you have two GPUs, you can choose a different\n    one as the default. Defaults to device 0 (the first device).\n\n  * `:preallocate`- if the memory should be preallocated on\n    GPU devices. Defaults to `true`.\n\n  * `:memory_fraction` - how much memory of a GPU device to\n    allocate. Defaults to `0.9`.","ref":"EXLA.html#module-client-options"},{"type":"module","title":"Memory preallocation - EXLA","doc":"XLA preallocates memory in GPU devices. This means that, if you are to\nrun multiple notebooks or multiple instances of your application, the\nsecond, third, and so on instances won't be able to allocate memory.\n\nYou can disable this behaviour by setting `preallocate: false` on the\nclient configuration, as specified above. You may also use\n`:memory_fraction` to control how much is preallocated.","ref":"EXLA.html#module-memory-preallocation"},{"type":"module","title":"GPU Runtime Issues - EXLA","doc":"GPU Executions run in dirty IO threads, which have a considerably smaller\nstack size than regular scheduler threads. This may lead to problems with\ncertain CUDA or cuDNN versions, leading to segmentation fails. In a development\nenvironment, it is suggested to set:\n\n    ELIXIR_ERL_OPTIONS=\"+sssdio 128\"\n\nTo increase the stack size of dirty IO threads from 40 kilowords to\n128 kilowords. In a release, you can set this flag in your `vm.args`.","ref":"EXLA.html#module-gpu-runtime-issues"},{"type":"module","title":"Distribution - EXLA","doc":"EXLA allows its tensors to be sent across nodes, as long as the parent\nnode (which effectively holds the tensor) keeps a reference to the\ntensor while it is read by any other node it was sent to.\n\nThe result of `EXLA.compile/3` can also be shared across nodes.\nOn invocation, the underlying executable is automatically serialized\nand sent to other nodes, without requiring a full recompilation,\nas long as the same conditions as above apply.","ref":"EXLA.html#module-distribution"},{"type":"module","title":"Docker considerations - EXLA","doc":"EXLA should run fine on Docker with one important consideration:\nyou must not start the Erlang VM as the root process in Docker.\nThat's because when the Erlang VM runs as root, it has to manage\nall child programs.\n\nAt the same time, Google's XLA shells out to child programs and\nmust retain control over how child programs terminate.\n\nTo address this, simply make sure you wrap the Erlang VM in\nanother process, such as a shell process. In other words, if you\nare using releases, instead of this:\n\n    CMD path/to/release start\n\ndo this:\n\n    CMD sh -c \"path/to/release start\"\n\nIf you are using Mix inside your Docker containers, instead of this:\n\n    CMD mix run\n\ndo this:\n\n    CMD sh -c \"mix run\"\n\nAlternatively, you can pass the `--init` flag to `docker run`,\nso it runs an `init` inside the container that forwards signals\nand reaps processes.\n\nThe `--init` flag uses the [`tini`](https://github.com/krallin/tini)\nproject, so for cases where the flag may not available (e.g.\nkubernetes) you may want to install it.","ref":"EXLA.html#module-docker-considerations"},{"type":"module","title":"Telemetry events - EXLA","doc":"EXLA executes a telemetry event every time a function is JIT-compiled.\nThe events are named `[:exla, :compilation]` and include the following\nmeasurements, given in microseconds:\n\n  * `:eval_time` - the time spent on turning the function into XLA\n    computation\n  * `:compile_time` - the time spent on compiling the XLA computation\n    into an executable\n  * `:total_time` - the sum of `:eval_time` and `:compile_time`\n\nThe metadata is:\n\n  * `:key` - the compilation key for debugging","ref":"EXLA.html#module-telemetry-events"},{"type":"function","title":"EXLA.cached?/3","doc":"Checks if the compilation of function with args is cached.\n\nNote that hooks are part of the cache, and\ntherefore they must be included in the options.","ref":"EXLA.html#cached?/3"},{"type":"function","title":"Examples - EXLA.cached?/3","doc":"iex> fun = fn a, b -> Nx.add(a, b) end\n    iex> left = Nx.tensor(1, type: {:u, 8})\n    iex> right = Nx.tensor([1, 2, 3], type: {:u, 16})\n    iex> EXLA.jit(fun).(left, right)\n    iex> EXLA.cached?(fun, [left, right])\n    true\n    iex> EXLA.cached?(fun, [left, Nx.tensor([1, 2, 3, 4], type: {:u, 16})])\n    false\n\nCompiled functions are also cached, unless cache is set to false:\n\n    iex> fun = fn a, b -> Nx.subtract(a, b) end\n    iex> left = Nx.tensor(1, type: {:u, 8})\n    iex> right = Nx.tensor([1, 2, 3], type: {:u, 16})\n    iex> EXLA.compile(fun, [left, right], cache: false)\n    iex> EXLA.cached?(fun, [left, right])\n    false\n    iex> EXLA.compile(fun, [left, right])\n    iex> EXLA.cached?(fun, [left, right])\n    true","ref":"EXLA.html#cached?/3-examples"},{"type":"function","title":"EXLA.compile/3","doc":"A shortcut for `Nx.Defn.compile/3` with the EXLA compiler.\n\n    iex> fun = EXLA.compile(&Nx.add(&1, &1), [Nx.template({3}, {:s, 32})])\n    iex> fun.(Nx.tensor([1, 2, 3]))\n    #Nx.Tensor \n\nThe returned function can be sent across nodes, as long as the parent\nnode (which effectively holds the function) keeps a reference to the\nfunction while it is invoked by any other node it was sent to. On\ninvocation, the underlying executable is automatically serialized\nand sent to other nodes, without requiring a full recompilation.\n\nSee `jit/2` for supported options.","ref":"EXLA.html#compile/3"},{"type":"function","title":"EXLA.jit/2","doc":"A shortcut for `Nx.Defn.jit/2` with the EXLA compiler.\n\n    iex> EXLA.jit(&Nx.add(&1, &1)).(Nx.tensor([1, 2, 3]))\n    #Nx.Tensor \n\nResults are allocated on the `EXLA.Backend`. Note that the\n`EXLA.Backend` is asynchronous: operations on its tensors\n*may* return immediately, before the tensor data is available.\nThe backend will then block only when trying to read the data\nor when passing it to another operation.","ref":"EXLA.html#jit/2"},{"type":"function","title":"Options - EXLA.jit/2","doc":"It accepts the same option as `Nx.Defn.jit/2` plus:\n\n  * `:cache` - cache the results of compilation, defaults to `true`.\n    You may disable it by setting it to `false`. You can also set it\n    to a binary, representing a filesystem path to store the cache.\n    EXLA will ensure the arguments and parameters across invocations\n    have the same shape, but it is ultimately your responsibility\n    to provide a unique cache path.\n\n  * `:client` - an atom representing the client to use. The default\n    client is chosen on this order: `:cuda`, `:rocm`, `:tpu`, and `:host`.\n\n  * `:debug` - print compile and debugging information, defaults to `false`.\n\n  * `:device_id` - the default device id to run the computation on.\n    Defaults to the `:default_device_id` on the client\n\n  * `:lazy_transfers` - when `:always`, it lazily transfers data to the device\n    instead of upfront. This is useful to reduce memory allocation on GPU/TPU\n    devices at the cost of increased latency. **It is recommended to only enable\n    this if the input tensors are allocated on host and the computation is\n    running on GPU/TPU with a limited amount of memory**","ref":"EXLA.html#jit/2-options"},{"type":"function","title":"EXLA.jit_apply/3","doc":"A shortcut for `Nx.Defn.jit_apply/3` with the EXLA compiler.\n\n    iex> EXLA.jit_apply(&Nx.add(&1, &1), [Nx.tensor([1, 2, 3])])\n    #Nx.Tensor \n\nSee `jit/2` for supported options.","ref":"EXLA.html#jit_apply/3"},{"type":"function","title":"EXLA.stream/3","doc":"Starts streaming the given anonymous function with just-in-time\ncompilation.\n\nAt least two arguments are expected:\n\n  1. The first argument is a tensor template of the data to\n     be streamed in\n\n  2. The second argument is a tensor with the stream initial state\n\nThe streaming function must return a two element tuple, the\nfirst element is the data to be sent and the second is the\naccumulator.\n\nFor each streamed chunk, you must call `Nx.Stream.send/2` and\n`Nx.Stream.recv/1`. You don't need to call `recv` immediately\nafter `send`, but doing so can be a useful mechanism to provide\nbackpressure. Once all chunks are sent, you must use `Nx.Stream.done/1`\nto receive the accumulated result. Let's see an example:\n\n    defmodule Streamed do\n      import Nx.Defn\n\n      defn sum(tensor, acc) do\n        {acc, tensor + acc}\n      end\n    end\n\nNow let's invoke it:\n\n    stream = EXLA.stream(&Streamed.sum/2, [Nx.template({}, {:s, 32}), 0])\n\n    for i <- 1..5 do\n      Nx.Stream.send(stream, i)\n      IO.inspect {:chunk, Nx.Stream.recv(stream)}\n    end\n\n    IO.inspect {:result, Nx.Stream.done(stream)}\n\nIt will print:\n\n    {:chunk, 0}\n    {:chunk, 1}\n    {:chunk, 2}\n    {:chunk, 3}\n    {:chunk, 4}\n    {:result, 5}\n\n**Note:** While any process can call `Nx.Stream.send/2`, EXLA\nexpects the process that starts the streaming to be the one\ncalling `Nx.Stream.recv/1` and `Nx.Stream.done/1`.\n\nSee `jit/2` for supported options.","ref":"EXLA.html#stream/3"},{"type":"function","title":"EXLA.stream_cached?/3","doc":"Checks if the JIT compilation of stream with\nargs is cached.\n\nNote that hooks are part of the cache, and\ntherefore they must be included in the options.","ref":"EXLA.html#stream_cached?/3"},{"type":"function","title":"Examples - EXLA.stream_cached?/3","doc":"iex> left = Nx.tensor(1, type: {:u, 8})\n    iex> right = Nx.tensor([1, 2, 3], type: {:u, 16})\n    iex> fun = fn x, acc -> {acc, Nx.add(x, acc)} end\n    iex> stream = EXLA.stream(fun, [left, right])\n    iex> Nx.Stream.done(stream)\n    iex> EXLA.stream_cached?(fun, [left, right])\n    true\n    iex> EXLA.stream_cached?(fun, [left, Nx.tensor([1, 2, 3, 4], type: {:u, 16})])\n    false","ref":"EXLA.html#stream_cached?/3-examples"},{"type":"function","title":"EXLA.to_mlir_module/3","doc":"Takes in a function, the argument templates and the compilation\noptions and returns the textual representation of the MLIR module.","ref":"EXLA.html#to_mlir_module/3"},{"type":"function","title":"Options - EXLA.to_mlir_module/3","doc":"* `:within_defn_compiler` - a boolean that indicates whether\n    this function is being called from within a `defn` compiler.\n    Defaults to `false`.","ref":"EXLA.html#to_mlir_module/3-options"},{"type":"function","title":"Examples - EXLA.to_mlir_module/3","doc":"iex> fun = fn x, y -> Nx.add(Nx.sin(x), Nx.cos(y)) end\n    iex> args = [1.0, 2.0]\n    iex> %{mlir_module: mlir_module} = EXLA.to_mlir_module(fun, args)\n    iex> mlir_module\n    \"\"\"\n    module {\n      func.func public @main(%arg0: tensor , %arg1: tensor ) -> tensor  {\n        %0 = stablehlo.sine %arg0 : tensor \n        %1 = stablehlo.cosine %arg1 : tensor \n        %2 = stablehlo.add %0, %1 : tensor \n        return %2 : tensor \n      }\n    }\n    \"\"\"","ref":"EXLA.html#to_mlir_module/3-examples"},{"type":"module","title":"EXLA.Backend","doc":"A Nx tensor backend for the data kept on the device.\n\nYou can directly transfer to this backend by calling\n`Nx.backend_transfer/2` or `Nx.backend_copy/2`. It\nallows the following options:\n\n  * `:client` - the client to store the data on.\n    Defaults to EXLA's default client.\n\n  * `:device_id` - which device to store it on.\n\nTo get the data out of the device backend into a regular\ntensor, call `Nx.backend_transfer/1` (with the device\ntensor as the single argument).\n\nNote that the `EXLA.Backend` is asynchronous: operations\non its tensors *may* return immediately, before the tensor\ndata is available. The backend will then block only when\ntrying to read the data or when passing it to another operation.","ref":"EXLA.Backend.html"},{"type":"module","title":"EXLA.BinaryBuffer","doc":"A buffer where data is kept in a binary.","ref":"EXLA.BinaryBuffer.html"},{"type":"function","title":"EXLA.BinaryBuffer.from_binary/2","doc":"","ref":"EXLA.BinaryBuffer.html#from_binary/2"},{"type":"module","title":"EXLA.Client","doc":"Functions for managing `EXLA.Client`.\n\nSee `EXLA` module docs for a general introduction.","ref":"EXLA.Client.html"},{"type":"function","title":"EXLA.Client.default_name/0","doc":"Returns the name of the default client.","ref":"EXLA.Client.html#default_name/0"},{"type":"function","title":"EXLA.Client.fetch!/1","doc":"Fetches a client with the given `name` from configuration.","ref":"EXLA.Client.html#fetch!/1"},{"type":"function","title":"EXLA.Client.from_outfeed/5","doc":"Sends buffer from device outfeed to the given process tagged by `ref`.","ref":"EXLA.Client.html#from_outfeed/5"},{"type":"function","title":"EXLA.Client.get_supported_platforms/0","doc":"Returns a map of supported platforms with device information.","ref":"EXLA.Client.html#get_supported_platforms/0"},{"type":"function","title":"EXLA.Client.to_infeed/3","doc":"Sends `data_and_typespecs` to device infeed.\n\n`data_and_typespecs` is a list of values corresponding to a single\ninfeed operation. It must be a list of two element tuples where the\nfirst element is a binary and the second element is a `EXLA.Typespec`.","ref":"EXLA.Client.html#to_infeed/3"},{"type":"module","title":"EXLA.DeviceBuffer","doc":"An EXLA DeviceBuffer for data allocated in the device.","ref":"EXLA.DeviceBuffer.html"},{"type":"function","title":"EXLA.DeviceBuffer.copy_to_device/3","doc":"Copies buffer to device with given device ID.","ref":"EXLA.DeviceBuffer.html#copy_to_device/3"},{"type":"function","title":"EXLA.DeviceBuffer.deallocate/1","doc":"Deallocates the underlying buffer.\n\nReturns `:ok` | `:already_deallocated`.","ref":"EXLA.DeviceBuffer.html#deallocate/1"},{"type":"function","title":"EXLA.DeviceBuffer.place_on_device/4","doc":"Places the given binary `data` on the given `device` using `client`.","ref":"EXLA.DeviceBuffer.html#place_on_device/4"},{"type":"function","title":"EXLA.DeviceBuffer.read/2","doc":"Reads `size` from the underlying buffer ref.\n\nThis copies the underlying device memory into a binary\nwithout destroying it. If `size` is negative, then it\nreads the whole buffer.","ref":"EXLA.DeviceBuffer.html#read/2"},{"type":"module","title":"EXLA.Executable","doc":"Wrapper around XLA's executable.","ref":"EXLA.Executable.html"},{"type":"function","title":"EXLA.Executable.dump/1","doc":"Dumps the executable to a data structure that can be serialized\nwith `term_to_binary`.\n\nWorks across nodes.","ref":"EXLA.Executable.html#dump/1"},{"type":"function","title":"EXLA.Executable.load/2","doc":"Loads a previously dumped executable.","ref":"EXLA.Executable.html#load/2"},{"type":"function","title":"EXLA.Executable.run/3","doc":"Runs the given executable with a list of lists as inputs and the given options.\n\nWorks across nodes.","ref":"EXLA.Executable.html#run/3"},{"type":"module","title":"EXLA.Typespec","doc":"Combined type and shape information about tensors.\n\nIn addition to the Nx types, also supports `{:pred, 8}` and `:token`,\nwhich are used internally within the compiler.\n\nThis struct corresponds to the `xla::Shape` class in the XLA compiler,\nbut is also meant as a lightweight data structure for passing the\ninformation around.\n\nNote: the name \"typespec\" has been chosen intentionally to distinguish\nit from both \"type\" and \"shape\".","ref":"EXLA.Typespec.html"},{"type":"function","title":"EXLA.Typespec.tensor/2","doc":"Builds a tensor typespec.","ref":"EXLA.Typespec.html#tensor/2"},{"type":"function","title":"EXLA.Typespec.to_shape/2","doc":"Returns an updated typespec with the given shape.","ref":"EXLA.Typespec.html#to_shape/2"},{"type":"function","title":"EXLA.Typespec.to_type/2","doc":"Returns an updated typespec with the given type.","ref":"EXLA.Typespec.html#to_type/2"},{"type":"function","title":"EXLA.Typespec.token/0","doc":"Builds a token typespec.","ref":"EXLA.Typespec.html#token/0"},{"type":"extras","title":"Rotating Image","doc":"# Rotating Image\n\n```elixir\nMix.install(\n  [\n    {:nx, github: \"elixir-nx/nx\", override: true, sparse: \"nx\"},\n    {:req, \"~> 0.3.5\"},\n    {:kino, \"~> 0.8.1\"},\n    {:exla, \"~> 0.5\"},\n    {:stb_image, \"~> 0.6\"}\n  ],\n  config: [\n    nx: [default_backend: EXLA.Backend]\n  ]\n)\n```","ref":"rotating-image.html"},{"type":"extras","title":"Preprocessing - Rotating Image","doc":"> Note: if you want to use TPUs/GPUs, you must set the XLA_TARGET environment variable.\n> See [`XLA_TARGET`](https://github.com/elixir-nx/xla#xla_target).\n\nPreprocessing a binary to the tensor representation and calculating shapes of the rotated image.\n\n```elixir\ndefmodule Preprocess do\n  def image_to_tensor(file_or_url) do\n    case URI.parse(file_or_url) do\n      %URI{scheme: nil} ->\n        img = StbImage.read_file!(file_or_url)\n        Nx.from_binary(img.data, {:u, 8}) |> Nx.reshape(img.shape)\n\n      _ ->\n        %{body: binary_image} = Req.get!(file_or_url)\n        img = StbImage.read_binary!(binary_image)\n        Nx.from_binary(img.data, {:u, 8}) |> Nx.reshape(img.shape)\n    end\n  end\n\n  defp calculate_new_side_length(side1, side2, angle, radians?) do\n    angle =\n      if radians?, do: angle, else: angle |> Nx.divide(180) |> Nx.multiply(Nx.Constants.pi())\n\n    sine = Nx.sin(angle)\n    cosine = Nx.cos(angle)\n\n    side1\n    |> Nx.multiply(cosine)\n    |> Nx.abs()\n    |> Nx.add(Nx.abs(Nx.multiply(side2, sine)))\n    |> Nx.add(1)\n    |> Nx.to_number()\n    |> round()\n  end\n\n  def get_new_height_and_width(height, width, angle, radians? \\\\ true) do\n    new_height = calculate_new_side_length(height, width, angle, radians?)\n    new_width = calculate_new_side_length(width, height, angle, radians?)\n    {new_height, new_width}\n  end\nend\n```","ref":"rotating-image.html#preprocessing"},{"type":"extras","title":"Rotating image - Rotating Image","doc":"```elixir\ndefmodule Rotate do\n  import Nx.Defn\n  @num_channels 4\n  # using three shears to avoid aliasing\n  # https://datagenetics.com/blog/august32013/index.html\n  defnp calculate_new_positions(\n          {y_image, x_image, _},\n          orig_ctr_height,\n          orig_ctr_width,\n          new_ctr_height,\n          new_ctr_width,\n          angle\n        ) do\n    k = Nx.iota({y_image * x_image})\n    i = Nx.as_type(k / x_image, {:u, 32})\n    j = Nx.remainder(k, x_image)\n\n    y = y_image - (orig_ctr_height + i + 1)\n\n    x = x_image - (orig_ctr_width + j + 1)\n\n    tangent = Nx.tan(angle / 2)\n\n    new_x = Nx.round(x - y * tangent)\n\n    new_y = Nx.round(Nx.sin(angle) * new_x + y)\n\n    new_x = Nx.round(new_x - new_y * tangent)\n\n    new_y = new_ctr_height - new_y\n    new_x = new_ctr_width - new_x\n\n    new_y = Nx.as_type(Nx.round(new_y), {:u, 32})\n    new_x = Nx.as_type(Nx.round(new_x), {:u, 32})\n\n    Nx.stack([new_y, new_x], axis: 1)\n  end\n\n  defnp calculate_ctr(coordinate) do\n    Nx.round((coordinate + 1) / 2 - 1)\n  end\n\n  defnp preprocess_position(pos, n) do\n    Nx.concatenate(\n      [\n        Nx.new_axis(pos, 1) |> Nx.tile([1, @num_channels, 1]),\n        Nx.iota({n, @num_channels, 1}, axis: 1)\n      ],\n      axis: 2\n    )\n    |> Nx.reshape({n * @num_channels, 3})\n  end\n\n  # new_image is a tensor with dims of rotated image filled with zeros\n\n  defn rotate_image_by_angle(image, angle, new_image, radians?) do\n    # if radians? is set to false than assuming that angle is given in degrees\n    angle = if radians?, do: angle, else: angle / 180 * Nx.Constants.pi()\n\n    {height, width, _} = image.shape\n    {new_height, new_width, _} = new_image.shape\n\n    orig_ctr_height = calculate_ctr(height)\n    orig_ctr_width = calculate_ctr(width)\n\n    new_ctr_height = calculate_ctr(new_height)\n    new_ctr_width = calculate_ctr(new_width)\n\n    pos =\n      calculate_new_positions(\n        image.shape,\n        orig_ctr_height,\n        orig_ctr_width,\n        new_ctr_height,\n        new_ctr_width,\n        angle\n      )\n\n    {n, 2} = Nx.shape(pos)\n\n    preprocessed_pos = preprocess_position(pos, n)\n    image = image |> Nx.reshape({n * @num_channels})\n\n    Nx.indexed_add(new_image, preprocessed_pos, image)\n  end\nend\n```","ref":"rotating-image.html#rotating-image"},{"type":"extras","title":"Example of usage - Rotating Image","doc":"```elixir\nnum_channels = 4\n\ndir = File.cwd!()\n\ndefault_input =\n  \"https://upload.wikimedia.org/wikipedia/commons/7/7f/Earth_fluent_design_reflective_surface_icon.png\"\n\ninput = Kino.Input.text(\"Image to rotate\", default: default_input)\nKino.render(input)\npath_to_file = Kino.Input.read(input)\n\n# set options\nradians? = 0\nangle = 90\n\n# rotate image by 70 degrees to the right\nimage = Preprocess.image_to_tensor(path_to_file)\n{height, width, _} = image.shape\n{new_height, new_width} = Preprocess.get_new_height_and_width(height, width, angle, radians?)\nnew_image = Nx.broadcast(Nx.tensor([0], type: {:u, 8}), {new_height, new_width, num_channels})\nrotate_image = EXLA.jit(&Rotate.rotate_image_by_angle/4)\nrotated_image = rotate_image.(image, angle, new_image, radians?)\n\n# convert the tensor back to stb_image and render it as png\nimg = StbImage.from_nx(rotated_image)\ncontent = StbImage.to_binary(img, :png)\n\n# display the rotated image\nKino.Image.new(content, :png)\n```","ref":"rotating-image.html#example-of-usage"},{"type":"extras","title":"Changelog","doc":"# Changelog","ref":"changelog.html"},{"type":"extras","title":"v0.9.0 (2024-09-26) - Changelog","doc":"","ref":"changelog.html#v0-9-0-2024-09-26"},{"type":"extras","title":"Enhancements - Changelog","doc":"* Overall improvements to the Nx.Defn compiler\n  * Compiled functions now work across BEAM nodes\n  * Add `cache: \"path/to/file\"` for disk caching JIT/compiled functions","ref":"changelog.html#enhancements"},{"type":"extras","title":"Bug fixes - Changelog","doc":"* Use a single thread pool for MLIR contexts","ref":"changelog.html#bug-fixes"},{"type":"extras","title":"v0.8.0 (2024-08-19) - Changelog","doc":"* Add `EXLA.to_mlir_module/2`\n  * The precompiled XLA CUDA binaries now require CUDA 12.1+ and cuDNN 9.1+\n  * Renamed `XLA_TARGET` value \"cuda120\" to \"cuda12\"\n  * `XLA_TARGET` automatically defaults to \"cuda12\" when CUDA installation is detected\n  * Allow NIF modules to be upgradable","ref":"changelog.html#v0-8-0-2024-08-19"},{"type":"extras","title":"v0.7.1 (2024-02-27) - Changelog","doc":"* Add CustomCallOp for QR decomposition\n  * Minor improvements to the MLIR modules generated\n  * MLIR Context pooling for better concurrency","ref":"changelog.html#v0-7-1-2024-02-27"},{"type":"extras","title":"v0.7.0 (2024-02-22) - Changelog","doc":"* Update to latest Nx\n  * Introduce a `:mlir` based compiler and use it by default. The previous `:xla` based compiler is deprecatead. You can temporarily revert to the previous compiler by setting `config :exla, :compiler_mode, :xla`","ref":"changelog.html#v0-7-0-2024-02-22"},{"type":"extras","title":"v0.6.4 (2023-11-13) - Changelog","doc":"* Update to latest Nx\n  * Allow `:automatic_transfers` configuration on client\n  * Do not discard client/device in `EXLA.Backend` when it is host\n  * Always sort NaN last\n  * Improve the `:axes` option in `gather`, `indexed_add`, and `indexed_put`","ref":"changelog.html#v0-6-4-2023-11-13"},{"type":"extras","title":"v0.6.3 (2023-11-09) - Changelog","doc":"* Update to latest Nx\n  * Fix mixed device usage on EXLA.Backend","ref":"changelog.html#v0-6-3-2023-11-09"},{"type":"extras","title":"v0.6.1 (2023-09-12) - Changelog","doc":"* Update to latest Nx","ref":"changelog.html#v0-6-1-2023-09-12"},{"type":"extras","title":"v0.6.0 (2023-08-15) - Changelog","doc":"* Allow cross-device transfers on host\n  * Update dependencies to OpenXLA\n  * Update to latest Nx","ref":"changelog.html#v0-6-0-2023-08-15"},{"type":"extras","title":"v0.5.3 (2023-04-14) - Changelog","doc":"* Fix compilation issue on certain macOS caused by O3\n  * Fix optimization which would cause EXLA to return a complete tuple instead of a subset","ref":"changelog.html#v0-5-3-2023-04-14"},{"type":"extras","title":"v0.5.2 (2023-03-21) - Changelog","doc":"* Automatically transfer tensors between nodes","ref":"changelog.html#v0-5-2-2023-03-21"},{"type":"extras","title":"v0.5.1 (2023-02-18) - Changelog","doc":"* Support `top_k`","ref":"changelog.html#v0-5-1-2023-02-18"},{"type":"extras","title":"v0.5.0 (2023-02-10) - Changelog","doc":"* Optimize backend_transfer/backend_copy within EXLA backends\n  * Use relative symlinks on compilation whenever possible","ref":"changelog.html#v0-5-0-2023-02-10"},{"type":"extras","title":"v0.4.2 (2023-01-13) - Changelog","doc":"","ref":"changelog.html#v0-4-2-2023-01-13"},{"type":"extras","title":"Enhancements - Changelog","doc":"* Automatically transfer from `:host` to other devices\n  * Support `lazy_transfers: :always` on `EXLA.jit/3` and `EXLA.compile/2`\n  * Run hooks concurrently once they have received the data","ref":"changelog.html#enhancements"},{"type":"extras","title":"Bug fixes - Changelog","doc":"* Respect default `EXLA.Backend` client when jitting argumentless operations\n  * Do not pick client without devices when loading initial client\n  * Consider the first conditional of a `cond` part of the current scope","ref":"changelog.html#bug-fixes"},{"type":"extras","title":"v0.4.1 (2022-12-07) - Changelog","doc":"","ref":"changelog.html#v0-4-1-2022-12-07"},{"type":"extras","title":"Enhancements - Changelog","doc":"* [EXLA] Require Nx ~> 0.4.1\n  * [EXLA] Update `XLA` to TF2.11\n  * [EXLA.Defn] Send telemetry event after XLA compilation\n  * [EXLA.Op] Add optimization barriers as operations","ref":"changelog.html#enhancements"},{"type":"extras","title":"Bug fixes - Changelog","doc":"* [EXLA] Validate backend options\n  * [EXLA.Backend] Fix `Nx.{any,all}` with `:keep_axes`\n  * [EXLA.Backend] Make SVD return `V` instead of `transpose(V)`\n  * [EXLA.Backend] Preserve NaNs in `window` and `reduce` operations","ref":"changelog.html#bug-fixes"},{"type":"extras","title":"v0.4.0 (2022-10-25) - Changelog","doc":"","ref":"changelog.html#v0-4-0-2022-10-25"},{"type":"extras","title":"Enhancements - Changelog","doc":"* Support zero copy binaries\n  * Redirect group leader for EXLA hooks","ref":"changelog.html#enhancements"},{"type":"extras","title":"Bug fixes - Changelog","doc":"* Always hoist `cond` expressions\n  * Fix conditional inside `Nx.map`","ref":"changelog.html#bug-fixes"},{"type":"extras","title":"v0.3.0 (2022-08-13) - Changelog","doc":"","ref":"changelog.html#v0-3-0-2022-08-13"},{"type":"extras","title":"Enhancements - Changelog","doc":"* Support `debug: true` option on `defn` compiler\n  * Allow specifying preferred clients via the application environment\n  * Support new callbacks added in Nx v0.3.0","ref":"changelog.html#enhancements"},{"type":"extras","title":"Deprecations - Changelog","doc":"* Deprecate `set_as_nx_default`","ref":"changelog.html#deprecations"},{"type":"extras","title":"v0.2.3 (2022-07-05) - Changelog","doc":"","ref":"changelog.html#v0-2-3-2022-07-05"},{"type":"extras","title":"Bug fixes - Changelog","doc":"* Fix predicate handling inside `cond`/`while`\n  * Set Nx backend globally","ref":"changelog.html#bug-fixes"},{"type":"extras","title":"v0.2.2 (2022-06-15) - Changelog","doc":"","ref":"changelog.html#v0-2-2-2022-06-15"},{"type":"extras","title":"Bug fixes - Changelog","doc":"* Fix invalid cache expiration when defn received functions as arguments","ref":"changelog.html#bug-fixes"},{"type":"extras","title":"v0.2.1 (2022-06-04) - Changelog","doc":"","ref":"changelog.html#v0-2-1-2022-06-04"},{"type":"extras","title":"Enhancements - Changelog","doc":"* Implement `EXLA.Backend.to_batched_list/3`","ref":"changelog.html#enhancements"},{"type":"extras","title":"Bug fixes - Changelog","doc":"* Improve support for non-finite values in `EXLA` compiler\n  * Fix segmentation fault while deallocating tensors","ref":"changelog.html#bug-fixes"},{"type":"extras","title":"v0.2.0 (2022-04-28) - Changelog","doc":"First release.","ref":"changelog.html#v0-2-0-2022-04-28"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,52,46,50]}}